{% if columns %}
with open('simpleimputer-numeric.pkl', 'rb') as f:
    __simple_imputer = pickle.load(f)

_NUMERIC_COLS_WITH_MISSING_VALUES = {{ columns }}
{{ test_dataset }}[_NUMERIC_COLS_WITH_MISSING_VALUES] = __simple_imputer.transform({{ test_dataset }}[_NUMERIC_COLS_WITH_MISSING_VALUES])
{% endif %}
{% if cols_almost_missing_numeric %}
_NUMERIC_ALMOST_MISSING_COLS = {{ cols_almost_missing_numeric }}
{{ test_dataset }}[_NUMERIC_ALMOST_MISSING_COLS] = {{ test_dataset }}[_NUMERIC_ALMOST_MISSING_COLS].fillna(0)
{% endif %}