{% if columns %}
with open('simpleimputer-string.pkl', 'rb') as f:
    __simple_imputer = pickle.load(f)

_STRING_COLS_WITH_MISSING_VALUES = {{ columns }}
{{ test_dataset }}[_STRING_COLS_WITH_MISSING_VALUES] = __simple_imputer.transform({{ test_dataset }}[_STRING_COLS_WITH_MISSING_VALUES])
{% endif %}
{% if cols_almost_missing_string %}
_STRING_ALMOST_MISSING_COLS = {{ cols_almost_missing_string }}
{{ test_dataset }}[_STRING_ALMOST_MISSING_COLS] = {{ test_dataset }}[_STRING_ALMOST_MISSING_COLS].fillna('')
{% endif %}